# Auto-Generated file - DO NOT EDIT!
# Source module: {{ module_full_name }}
# Generated at: {{ iso_timestamp }}
from __future__ import annotations
__all__ = [
{% for a in all_list %}
    '{{ a }}',
{% endfor %}
]
import dataclasses
import datetime
import enum
from typing import *
from protoplasm.casting import dictators
from protoplasm import plasm
{{ imports }}
{{ proto_import }}

import logging
log = logging.getLogger(__name__)


{% if enum_list %}
{% for enum in enum_list %}
{{ enum }}

{% endfor %}
{% endif %}
{% if message_list %}
{% for msg in message_list %}
{{ msg }}

{% endfor -%}
{%- endif -%}