# Auto-Generated file - DO NOT EDIT!
# Source module: {{ module_full_name }}
# Generated at: {{ iso_timestamp }}
__all__ = [
{% for a in all_list %}
    '{{ a }}',
{% endfor %}
]
from typing import *
from protoplasm import plasm
{{ imports }}
if TYPE_CHECKING:
    from grpc import ServicerContext
    {{ api_import }}

import logging
log = logging.getLogger(__name__)


{{ services }}